/**
 * Copyright (C) 2003-2018, Foxit Software Inc..
 * All Rights Reserved.
 *
 * http://www.foxitsoftware.com
 *
 * The following code is copyrighted and is the proprietary of Foxit Software Inc.. It is not allowed to
 * distribute any parts of Foxit Mobile PDF SDK to third party or public without permission unless an agreement
 * is signed between Foxit Software Inc. and customers to explicitly grant customers permissions.
 * Review legal.txt for additional license and legal information.
 */


#include "pch.h"
#include "MRUManager.h"

using namespace CompletePDFViewer;

using namespace Platform;
using namespace Windows::Storage;
using namespace Windows::Storage::Streams;
using namespace Windows::Storage::AccessCache;

#define MRU_CATALOG_NAME "MRU"
#define THUMBNAIL_NAME "Thumbnail"

MRUManager^ _Current;

ThumbnailFileSaveCompletionEventArgs::ThumbnailFileSaveCompletionEventArgs(Platform::String^ token)
{
	_Token = token;
}

Platform::String^ ThumbnailFileSaveCompletionEventArgs::Token::get()
{
	return _Token;
}

MRUManager::MRUManager()
{
	InitMRUManager();
}

void MRUManager::InitMRUManager()
{
	m_MRUFolder = nullptr;

	concurrency::create_task(ApplicationData::Current->LocalFolder->CreateFolderAsync(MRU_CATALOG_NAME, Windows::Storage::CreationCollisionOption::OpenIfExists))
		.then([this](Windows::Storage::StorageFolder^ folder)
	{
		m_MRUFolder = folder;
	});
}

MRUManager^ MRUManager::Current::get()
{
	if (_Current == nullptr)
	{
		_Current = ref new MRUManager();
	}

	return _Current;
}

bool MRUManager::IsStorageFileInMRUList(Windows::Storage::StorageFile^ file)
{
	return IsStorageFileInMRUList(file->Path);
}

Platform::String^ MRUManager::GetStorageFileMRUToken(Windows::Storage::StorageFile^ file)
{
	return GetStorageFileMRUToken(file->Path);
}

bool MRUManager::IsStorageFileInMRUList(Platform::String^ filepath)
{
	AccessListEntryView^ _entries = StorageApplicationPermissions::MostRecentlyUsedList->Entries;
	bool _isExist = false;

	auto _iter = _entries->First();
	while (_iter->HasCurrent)
	{
		auto _value = _iter->Current;
		if (_value.Metadata == filepath)
		{
			_isExist = true;
			break;
		}
		_iter->MoveNext();
	}
	return _isExist;
}

Platform::String^ MRUManager::GetStorageFileMRUToken(Platform::String^ filepath)
{
	AccessListEntryView^ _entries = StorageApplicationPermissions::MostRecentlyUsedList->Entries;

	auto _iter = _entries->First();
	while (_iter->HasCurrent)
	{
		auto _value = _iter->Current;
		if (_value.Metadata == filepath)
			return _value.Token;
		_iter->MoveNext();
	}
	return nullptr;
}

bool MRUManager::IsMatchTokenWithFilePath(Platform::String^ token, Platform::String^ filepath)
{
	bool _ismatch = false;
	AccessListEntryView^ _entries = StorageApplicationPermissions::MostRecentlyUsedList->Entries;

	if (StorageApplicationPermissions::MostRecentlyUsedList->ContainsItem(token))
	{
		auto _iter = _entries->First();
		while (_iter->HasCurrent)
		{
			auto _value = _iter->Current;
			if (_value.Token == token && _value.Metadata == filepath)
			{
				_ismatch = true;
				break;
			}
			_iter->MoveNext();
		}
	}
	
	return _ismatch;
}

Platform::String^ MRUManager::AddStorageFileToMRUList(Windows::Storage::StorageFile^ file)
{
	auto _token = GetStorageFileMRUToken(file);
	if (!_token)
	{
		_token = StorageApplicationPermissions::MostRecentlyUsedList->Add(file, file->Path);
		
		OutputDebugString(L"Add to RecentlyUseedList....\n");
	}
	return _token;
}

void MRUManager::DeleteFromMRUListByStorageFile(Windows::Storage::StorageFile^ file)
{
	DeleteFromMRUListByFilePath(file->Path);
}

void MRUManager::DeleteFromMRUListByFilePath(Platform::String^ filepath)
{
	DeleteFromMRUListByToken(GetStorageFileMRUToken(filepath));
}

void MRUManager::DeleteFromMRUListByToken(Platform::String^ token)
{
	if (token)
	{
		if (StorageApplicationPermissions::MostRecentlyUsedList->ContainsItem(token))
		{
			StorageApplicationPermissions::MostRecentlyUsedList->Remove(token);

			DeleteMRUThumbnailFile(token).then([](concurrency::task<void> t)
			{
				try
				{
					t.get();
				}
				catch (...)
				{
					OutputDebugString(L"DeleteMRUThumbnailFile...Failed...\n");
				}
			});

			if (ApplicationData::Current->LocalSettings->Values->HasKey(token))
			{
				ApplicationData::Current->LocalSettings->Values->Remove(token);
			}
		}
	}
}

Windows::Foundation::IAsyncOperation<Windows::UI::Xaml::Media::Imaging::WriteableBitmap^>^ MRUManager::LoadWriteableBitmapFromMRUThumbnailFileAsync(Platform::String^ token)
{
	return concurrency::create_async([=]()
	{
		return LoadWriteableBitmapFromMRUThumbnailFile(token)
			.then([=](Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ result) -> Windows::UI::Xaml::Media::Imaging::WriteableBitmap^
		{
			return result;
		});
	});
}

concurrency::task<void> MRUManager::LoadStreamFromWriteableBitmap(Windows::Storage::Streams::IRandomAccessStream^ stream, Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ bitmap)
{
	std::shared_ptr<DataWriter^> _spDataWriter = std::make_shared<DataWriter^>(nullptr);
	std::shared_ptr<DataReader^> _spDataReader = std::make_shared<DataReader^>(nullptr);

	IBuffer^ buffer = bitmap->PixelBuffer;
	DataReader^ dataReader = DataReader::FromBuffer(buffer);
	Array<unsigned char, 1>^ fileContent = ref new Array<unsigned char, 1>(buffer->Length);
	dataReader->ReadBytes(fileContent);

	IOutputStream^ outStream = stream->GetOutputStreamAt(0);
	*_spDataWriter = ref new DataWriter(outStream);
	*_spDataReader = ref new DataReader(stream);

	return concurrency::create_task(Windows::Graphics::Imaging::BitmapEncoder::CreateAsync(Windows::Graphics::Imaging::BitmapEncoder::JpegEncoderId, stream))
		.then([_spDataWriter, bitmap, fileContent, _spDataReader](Windows::Graphics::Imaging::BitmapEncoder^ encoder)
	{
		encoder->SetPixelData(Windows::Graphics::Imaging::BitmapPixelFormat::Bgra8,
			Windows::Graphics::Imaging::BitmapAlphaMode::Straight, bitmap->PixelWidth, bitmap->PixelHeight, 96, 96, fileContent);
		return encoder->FlushAsync();
	})
		.then([_spDataWriter, _spDataReader]()
	{
		IBuffer^ buf = (*_spDataReader)->ReadBuffer(0);
		(*_spDataWriter)->WriteBuffer(buf);
		return (*_spDataWriter)->StoreAsync();
	})
		.then([_spDataWriter, _spDataReader, outStream](unsigned int bytesStored)
	{
		(*_spDataWriter)->DetachStream();
		(*_spDataReader)->DetachStream();
		return outStream->FlushAsync();
	})
		.then([=](bool flushSucceeded)
	{
		OutputDebugString(L"LoadStreamFromWriteableBitmap --- FlushAsync...\n");
	});
}

concurrency::task<void> MRUManager::SaveMRUItemThumbnailToFileByToken(Platform::String^ token, Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ bitmap)
{
	return concurrency::create_task(ApplicationData::Current->LocalFolder->CreateFolderAsync(MRU_CATALOG_NAME, Windows::Storage::CreationCollisionOption::OpenIfExists))
		.then([token, bitmap](Windows::Storage::StorageFolder^ folder)
	{
		return folder->CreateFolderAsync(token, Windows::Storage::CreationCollisionOption::OpenIfExists);
	})
		.then([token, bitmap](Windows::Storage::StorageFolder^ folder)
	{
		return folder->CreateFileAsync(THUMBNAIL_NAME, Windows::Storage::CreationCollisionOption::ReplaceExisting);
	})
		.then([token, bitmap](Windows::Storage::StorageFile^ file)
	{
		return file->OpenAsync(FileAccessMode::ReadWrite);
	})
		.then([token, bitmap](Windows::Storage::Streams::IRandomAccessStream^ stream)
	{
		return LoadStreamFromWriteableBitmap(stream, bitmap);
	})
		.then([token]()
	{
		auto callBackUI = [token]()
		{
			auto args = ref new ThumbnailFileSaveCompletionEventArgs(token);
			MRUManager::Current->ThumbnailFileSaveCompletion(MRUManager::Current, args);
		};

		Windows::UI::Xaml::Window::Current->Dispatcher->RunAsync(Windows::UI::Core::CoreDispatcherPriority::Low, ref new Windows::UI::Core::DispatchedHandler(callBackUI));
	})
		.then([](concurrency::task<void> t)
	{
		try
		{
			t.get();
		}
		catch (Platform::Exception^ e)
		{
			OutputDebugString(e->Message->Data());
		}
	});
}

concurrency::task<void> MRUManager::SaveMRUItemThumbnailToFileByStorageFile(Windows::Storage::StorageFile^ file, Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ bitmap)
{
	auto _token = GetStorageFileMRUToken(file->Path);
	if (!_token)
	{
		_token = AddStorageFileToMRUList(file);
	}
	return SaveMRUItemThumbnailToFileByToken(_token, bitmap);
}


concurrency::task<Windows::UI::Xaml::Media::Imaging::WriteableBitmap^> MRUManager::LoadWriteableBitmapFromMRUThumbnailFile(Platform::String^ token)
{
	std::shared_ptr<Windows::UI::Xaml::Media::Imaging::WriteableBitmap^> _spWriteableBitmap = std::make_shared<Windows::UI::Xaml::Media::Imaging::WriteableBitmap^>(nullptr);

	return concurrency::create_task(ApplicationData::Current->LocalFolder->CreateFolderAsync(MRU_CATALOG_NAME, Windows::Storage::CreationCollisionOption::OpenIfExists))
		.then([token](Windows::Storage::StorageFolder^ folder)
	{
		return folder->CreateFolderAsync(token, Windows::Storage::CreationCollisionOption::OpenIfExists);
	})
		.then([token](Windows::Storage::StorageFolder^ folder)
	{
		return folder->TryGetItemAsync(THUMBNAIL_NAME);
	})
		.then([=](concurrency::task<Windows::Storage::IStorageItem^> t)
	{
		Windows::Storage::IStorageItem^ sampleItem = t.get();
		Windows::Foundation::IAsyncOperation<IRandomAccessStream^>^ op = Concurrency::create_async([]() ->IRandomAccessStream^ {  
			return (IRandomAccessStream^)nullptr;
		});  
		if(sampleItem == nullptr)
			return op;
		IStorageFile^ sampleFile = sampleItem->IsOfType(Windows::Storage::StorageItemTypes::File)?((IStorageFile^)sampleItem):nullptr;
		if(sampleFile == nullptr) 
			return op;
		return sampleFile->OpenAsync(FileAccessMode::Read);
	})
		.then([=](IRandomAccessStream^ readStream)
	{

		if (readStream == nullptr)
		{
			Windows::Foundation::IAsyncAction^ action = Concurrency::create_async([](){
			});
			return action;
		}
		auto callBackUI = [readStream, _spWriteableBitmap]()
		{
				auto _GeneralSetting = GeneralSetting::Current;
				int nWidth = (int)_GeneralSetting->MRUThumbnailWidth;
				int nHeight = (int)_GeneralSetting->MRUThumbnailHeight;
				Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ bitmapSnap = ref new Windows::UI::Xaml::Media::Imaging::WriteableBitmap(nWidth, nHeight);
				readStream->Seek(0);
				try {
					bitmapSnap->SetSource(readStream);
				}
				catch (...)
				{
					*_spWriteableBitmap = nullptr;
					return;
				}
				*_spWriteableBitmap = bitmapSnap;

		};

		return Windows::UI::Xaml::Window::Current->Dispatcher->RunAsync(Windows::UI::Core::CoreDispatcherPriority::Normal, ref new Windows::UI::Core::DispatchedHandler(callBackUI));
	})
		.then([_spWriteableBitmap]()
	{
		return *_spWriteableBitmap;
	})
		.then([](concurrency::task<Windows::UI::Xaml::Media::Imaging::WriteableBitmap^> t)
	{
		Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ ret = nullptr;
		try
		{
			ret = t.get();
		}
		catch (Platform::Exception^ e)
		{
			OutputDebugString(e->Message->Data());
		}
		return ret;
	});
}

concurrency::task<void> MRUManager::DeleteMRUThumbnailFile(Platform::String^ token)
{
	return concurrency::create_task(ApplicationData::Current->LocalFolder->CreateFolderAsync(MRU_CATALOG_NAME, Windows::Storage::CreationCollisionOption::OpenIfExists))
		.then([token](Windows::Storage::StorageFolder^ folder)
	{
		return folder->GetFolderAsync(token);
	})
		.then([token](Windows::Storage::StorageFolder^ folder)
	{
		return folder->DeleteAsync();
	});
}

concurrency::task<void> MRUManager::ConvertToImageFile(Platform::String^ FileName, Platform::String^ ImageFileName, Windows::UI::Xaml::Media::Imaging::WriteableBitmap^ bitmap)
{
	return concurrency::create_task(ApplicationData::Current->LocalFolder->CreateFolderAsync(FileName, Windows::Storage::CreationCollisionOption::OpenIfExists))
		.then([ImageFileName, bitmap](Windows::Storage::StorageFolder^ folder)	
	{
		return folder->CreateFileAsync(ImageFileName, Windows::Storage::CreationCollisionOption::ReplaceExisting);
	})
		.then([ImageFileName, bitmap](Windows::Storage::StorageFile^ file)
	{
		return file->OpenAsync(FileAccessMode::ReadWrite);
	})
		.then([ImageFileName, bitmap](Windows::Storage::Streams::IRandomAccessStream^ stream)
	{
		return LoadStreamFromWriteableBitmap(stream, bitmap);
	})
		.then([ImageFileName]()
	{
		
	});
}
