/**
 * Copyright (C) 2003-2018, Foxit Software Inc..
 * All Rights Reserved.
 *
 * http://www.foxitsoftware.com
 *
 * The following code is copyrighted and is the proprietary of Foxit Software Inc.. It is not allowed to
 * distribute any parts of Foxit Mobile PDF SDK to third party or public without permission unless an agreement
 * is signed between Foxit Software Inc. and customers to explicitly grant customers permissions.
 * Review legal.txt for additional license and legal information.
 */


#include "pch.h"
#include "DocUndo.h"
#include "Undo.h"

#include <assert.h>

using namespace foxit::rdk::uiextensions;

using namespace Platform;

DocUndo::DocUndo(int nBufsize) : m_nCurUndoPos(0),
	m_nBufSize(nBufsize),
	m_bModified(false),
	m_bVirgin(true),
	m_bWorking(false)
{
	m_nCurUndoPos = 0;
	m_UndoItems = ref new Platform::Collections::Vector<IUndoItem^>();
}

bool DocUndo::CanUndo()
{
	return m_nCurUndoPos > 0;
}

void DocUndo::Undo()
{
	m_bWorking = true;
	if (m_nCurUndoPos > 0)
	{
		IUndoItem^ pItem = m_UndoItems->GetAt(m_nCurUndoPos - 1);
		assert(pItem != nullptr);

		pItem->Undo();

		m_nCurUndoPos--;
		m_bModified = (m_nCurUndoPos != 0);
	}

	m_bWorking = false;
}

bool DocUndo::CanRedo()
{
	return m_nCurUndoPos < (int)m_UndoItems->Size;
}

void DocUndo::Redo()
{
	m_bWorking = true;

	unsigned int nStackSize = m_UndoItems->Size;

	if (m_nCurUndoPos < (int)nStackSize)
	{
		IUndoItem^ pItem = m_UndoItems->GetAt(m_nCurUndoPos);
		assert(pItem != nullptr);

		pItem->Redo();

		m_nCurUndoPos++;
		m_bModified = (m_nCurUndoPos != 0);
	}

	m_bWorking = false;
}

bool DocUndo::IsModified()
{
	if (m_bVirgin)
		return m_bModified;
	else
		return true;
}

bool DocUndo::IsWorking()
{
	return m_bWorking;
}

int DocUndo::GetItemCount()
{
	return m_UndoItems->Size;
}

int DocUndo::GetCurUndoPos()
{
	return m_nCurUndoPos;
}

void DocUndo::SetBufSize(int nSize)
{
	m_nBufSize = nSize;
}

int DocUndo::GetBufSize()
{
	return m_nBufSize;
}

void DocUndo::RemoveHeads()
{
	assert(m_UndoItems->Size > 1);

	IUndoItem^ pItem = m_UndoItems->GetAt(0);
	assert(pItem != nullptr);

	pItem->ReleaseItem();
	m_UndoItems->RemoveAt(0);
}

void DocUndo::RemoveTails()
{
	unsigned int _itemSize = m_UndoItems->Size;
	if (_itemSize == 0)
		return;

	for (int i = _itemSize - 1; i >= m_nCurUndoPos; i--)
	{
		IUndoItem^ pItem = m_UndoItems->GetAt(i);
		assert(pItem != nullptr);

		pItem->ReleaseItem();
		m_UndoItems->RemoveAt(i);
	}
}

void DocUndo::Reset()
{
	unsigned int nStackSize = m_UndoItems->Size;
	for (unsigned int i = 0, sz = nStackSize; i < sz; i++)
	{
		IUndoItem^ pItem = m_UndoItems->GetAt(i);
		assert(pItem != nullptr);

		pItem->ReleaseItem();
	}

	m_nCurUndoPos = 0;
	m_UndoItems->Clear();
}

void DocUndo::AddItem(IUndoItem^ pItem)
{
	assert(!m_bWorking);
	assert(pItem != nullptr);
	assert(m_nBufSize > 1);

	if (m_nCurUndoPos < (int)m_UndoItems->Size)
		RemoveTails();

	if ((int)m_UndoItems->Size >= m_nBufSize)
	{
		RemoveHeads();
		m_bVirgin = FALSE;
	}

	m_UndoItems->Append(pItem);
	m_nCurUndoPos = m_UndoItems->Size;

	m_bModified = (m_nCurUndoPos != 0);
}

IUndoItem^ DocUndo::GetItem(int nIndex)
{
	if (nIndex >= 0 && nIndex < (int)m_UndoItems->Size)
		return m_UndoItems->GetAt(nIndex);

	return nullptr;
}

void DocUndo::DeleteItem(IUndoItem^ pItem)
{
	int nCount = m_UndoItems->Size;
	if (nCount <= 0) return;

	unsigned int nIndex = -1;
	if (m_UndoItems->IndexOf(pItem, &nIndex))
	{
		IUndoItem^ pItem = m_UndoItems->GetAt(nIndex);
		assert(pItem != nullptr);

		pItem->ReleaseItem();
		m_UndoItems->RemoveAt(nIndex);
	}

	m_nCurUndoPos = m_UndoItems->Size;
}
